searchNodes=[{"doc":"Definition of boundaries within the application. A boundary is a named group of modules which can export some of its modules, and depend on other boundaries. Boundary definitions can be used in combination with Mix.Tasks.Compile.Boundary to restrain cross-module dependencies. A few examples of what you can do with boundary include: Prevent invocations from the context layer to the web layer Prevent invocations from the web layer to internal context modules Prevent usage of Phoenix and Plug in the context layer Limit usage of Ecto in the web layer to only Ecto.Changeset Allow :mix modules to be used only at compile time Quick example The following code defines boundaries for a typical Phoenix based project generated with mix phx.new. defmodule MySystem do use Boundary, deps: [], exports: [] # ... end defmodule MySystemWeb do use Boundary, deps: [MySystem], exports: [Endpoint] # ... end defmodule MySystem.Application do use Boundary, deps: [MySystem, MySystemWeb] # ... end These boundaries specify the allowed cross-boundary usage: Modules from MySystemWeb may use the MySystem module, but not other MySystem.* modules. MySystem.Application code may use MySystem, MySystemWeb, and MySystemWeb.Endpoint modules. Defining a boundary A boundary is defined via use Boundary expression in the top-level (aka root) module. For example, the context boundary named MySystem can be defined as follows: defmodule MySystem do use Boundary, opts # ... end Module classification Based on the existing definitions, modules are classified into boundaries. Each module can belong to at most one boundary. A module doesn&#39;t need to belong to a boundary, in which case we say that the module is unclassified. Boundary membership is determined from the module name. In the previous example, we defined a single boundary, called MySystem. This boundary will contain the root module (MySystem), as well as all modules whose name starts with MySystem.. In addition, it&#39;s possible to extract some of the modules from a boundary into its own boundary. For example: defmodule MySystem do use Boundary, opts end defmodule MySystem.Endpoint do use Boundary, opts end Here, modules from MySystem.Endpoint namespace are promoted into its own boundary. It&#39;s worth noting that MySystem.Endpoint is considered as a peer boundary of MySystem, not its child. At the moment, nesting of boundaries (defining internal boundaries within other boundaries) is not supported by this library. Mix tasks By convention, mix tasks have to reside in the Mix.Tasks namespace, which makes it harder to put them under the same boundary. To assist with this, boundary supports manual reclassification of such modules. The advised approach is to introduce the MySystem.Mix boundary which can hold helper functions required by the mix tasks. With such boundary in place, you can manually classify mix tasks as: defmodule Mix.Tasks.SomeTask do use Boundary, classify_to: MySystem.Mix use Mix.Task end defmodule Mix.Tasks.AnotherTask do use Boundary, classify_to: MySystem.Mix use Mix.Task end This way, both modules will be considered as a part of the MySystem.Mix boundary. Note that manual classification is allowed only for mix tasks and protocol implementations (see the following section). Protocol implementation Consider the following protocol implementation: defimpl String.Chars, for: MySchema, do: # ... This code will generate the module String.Chars.MySchema. Therefore, the module sits in a completely different &quot;namespace&quot;. In addition, the desired boundary of such module can vary from one case to another. In some cases, a protocol implementation might be a UI concern, while in others, it might be a domain concern. For these reasons, protocol implementations are treated in a special way. A protocol implementation is by default unclassified (it doesn&#39;t belong to any boundary). However, the boundary checker will not emit a warning for unclassified protocol implementations. However, you can manually classify the protocol implementation, as demonstrated in the following example: defimpl String.Chars, for: MySchema do use Boundary, classify_to: MySystem # ... end Note that :classify_to option is only allowed for protocol implementations and mix tasks. Other modules can&#39;t be manually classified. Exports Exports are boundary modules which can be used by modules from other boundaries. A boundary always exports its root module, and it may additionally export other modules, which can be configured with the :exports option: defmodule MySystem do use Boundary, exports: [User] end In this example, we&#39;re defining the MySystem boundary which exports the modules MySystem and MySystem.User. All other modules of this boundary are considered to be internal, and they may not be used by modules from other boundaries. Dependencies Each boundary may depend on other boundaries. These dependencies are used to defined allowed cross-boundary module usage. A module from another boundary may only be used if: The callee boundary is a direct dependency of the caller boundary. The callee boundary exports the used module. For example: defmodule MySystem do use Boundary, exports: [User], deps: [] end defmodule MySystemWeb do use Boundary, exports: [], deps: [MySystem] end In this example we specify the following rules: Code from the MySystem boundary can&#39;t use any module from other boundaries. Code from the MySystemWeb boundary may use exports of the MySystem boundary (MySystem and MySystem.User). Of course, in-boundary cross-module dependencies are always allowed (any module may use all other modules from the same boundary). When listing deps and exports, a &quot;grouping&quot; syntax can also be used: use Boundary, deps: [Foo.{Bar, Baz}] External dependencies By default, all dependencies on modules from other OTP applications are permitted. However, you can restrain such dependencies by including boundaries from the external application. For example, let&#39;s say you want to limit the ecto usage in the web tier to only Ecto.Changeset. This can be specified as follows: defmodule MySystemWeb do use Boundary, deps: [Ecto.Changeset] end Boundary is able to use boundary definitions from an external application, if such exists. If an external application doesn&#39;t define any boundary, you can still reference application modules. In such case, you&#39;re creating an implicit boundary. This is exactly what we&#39;re doing in the previous example. Ecto doesn&#39;t define its own boundaries, but we can still include Ecto.Changeset in the deps list. This will create an implicit boundary of the same name which will include all of the submodules like Ecto.Changeset.Foo, or Ecto.Changeset.Bar.Baz. An implicit boundary exports all of its submodules. Note that you can&#39;t define implicit boundaries in applications which define their own boundaries. The implicit boundaries are collected based on all deps of all boundaries in your application. For example, if one boundary specifies Ecto.Query as a dependency, while another references Ecto.Query.API, then two boundaries are defined, and the latter will not be a part of the former. In some cases you may want to completely prohibit the usage of some library. However, bare in mind that by default calls to an external application are restricted only if the client boundary references at least one dep boundary from that application. To force boundary to always restrict calls to some application, you can include the application in the :extra_externals list: defmodule MySystem do use Boundary, extra_externals: [:plug], deps: [] end The :extra_externals list contains additional applications which are always considered. Any calls to given applications must be explicitly allowed via the :deps option. In the example above, we&#39;re including :plug in the list of external applications, but we&#39;re not including any boundary from this library in depssi, the context layer is not allowed to use plug functions. In addition, a strict external mode is supported via the :externals_mode option: defmodule MySystem do use Boundary, externals_mode: :strict end In this mode, boundary will report all calls to all external applications which are not explicitly allowed via the :dep option. You can also configure the strict mode globally in your mix.exs: defmodule MySystem.MixProject do use Mix.Project def project do [ # ... boundary: [externals_mode: :strict] ] end # ... end At this point, all boundaries will be checked with the strict mode. If you want to override this for some boundaries, you can do it with use Boundary, externals_mode: :relaxed. Note that restraining calls to the :elixir, :boundary, and pure Erlang applications, such as :crypto or :cowboy, is currently not possible. If you want to discover which external applications are used by your boundaries, you can use the helper mix task Mix.Tasks.Boundary.FindExternalDeps. Compile-time dependencies By default, a dependency allows calls at both, compile time and runtime. In some cases you may want to permit calls to some dependency only at compile-time. A typical example are modules from the :mix application. These modules are not safe to be used at runtime. Limiting their usage to compile-time only can be done as follows: use Boundary, deps: [{Mix, :compile}] With such configuration, the following calls are allowed: Function invocations at compile time (i.e. outside of any function, or in unquote(...)). Macro invocations anywhere in the code. Note that you might have some modules which will require runtime dependency on mix, such as custom mix tasks. It&#39;s advised to group such modules under a common boundary, such as MySystem.Mix, and allow mix as a runtime dependency only in that boundary. Finally, it&#39;s worth noting that it&#39;s not possible to permit some dependency only at runtime. If a dependency is allowed at runtime, then it can also be used at compile time. Ignored boundaries It is possible to exclude some modules from cross-boundary checks by defining an ignored boundary: defmodule MySystem do use Boundary, ignore?: true end When a boundary is ignored, all modules belonging to it can use any other module, and can be used by any other module. The purpose of this option is to support relaxing rules in some parts of your code. For example, you may wish to ignore boundary constraints for your test support modules. By introducing a top-level boundary for such modules (e.g. MySystemTest), and marking this boundary as ignored, you can easily achieve that. Another scenario is when introducing boundaries in an existing, possibly large project, which has many complex dependencies that can&#39;t be untangled trivially. In such case, ignored boundaries provide a mechanism for gradually introducing boundaries into the project. For example, you could first define ignored boundaries which encompass the entire system: defmodule MySystem do use Boundary, ignore?: true end defmodule MySystemWeb do use Boundary, ignore?: true end Now, you can pick smaller parts of your code where you can clean up the dependencies: defmodule MySystem.Context1 do use Boundary, exports: [...], deps: [] end defmodule MySystemWeb.Controller1 do use Boundary, exports: [], deps: [MySystem.Context1] end Going further, you can gradually expand the parts of your code covered by non-ignored boundaries. Once you&#39;re properly covering the entire system, you can remove the intermediate finer-grained boundaries, and specify the rules at the higher-level: defmodule MySystem do use Boundary, exports: [...], deps: [] end defmodule MySystemWeb do use Boundary, exports: [Endpoint], deps: [...] end","ref":"Boundary.html","title":"Boundary","type":"module"},{"doc":"Returns definitions of all boundaries. The result will include boundaries from listed externals, as well as implicit boundaries. You shouldn&#39;t access the data in this result directly, as it may change significantly without warnings. Use exported functions of this module to acquire the information you need.","ref":"Boundary.html#all/1","title":"Boundary.all/1","type":"function"},{"doc":"Returns the names of all boundaries. The result will include boundaries from listed externals, as well as implicit boundaries.","ref":"Boundary.html#all_names/1","title":"Boundary.all_names/1","type":"function"},{"doc":"Returns the application of the given module.","ref":"Boundary.html#app/2","title":"Boundary.app/2","type":"function"},{"doc":"Returns all boundary errors.","ref":"Boundary.html#errors/2","title":"Boundary.errors/2","type":"function"},{"doc":"Returns the definition of the given boundary.","ref":"Boundary.html#fetch/2","title":"Boundary.fetch/2","type":"function"},{"doc":"Returns the definition of the given boundary.","ref":"Boundary.html#fetch!/2","title":"Boundary.fetch!/2","type":"function"},{"doc":"Returns definition of the boundary to which the given module belongs.","ref":"Boundary.html#for_module/2","title":"Boundary.for_module/2","type":"function"},{"doc":"Returns the definition of the given boundary.","ref":"Boundary.html#get/2","title":"Boundary.get/2","type":"function"},{"doc":"Returns true if the module is an implementation of some protocol.","ref":"Boundary.html#protocol_impl?/1","title":"Boundary.protocol_impl?/1","type":"function"},{"doc":"Returns the collection of unclassified modules.","ref":"Boundary.html#unclassified_modules/1","title":"Boundary.unclassified_modules/1","type":"function"},{"doc":"Builds the boundary-specific view of the given application.","ref":"Boundary.html#view/1","title":"Boundary.view/1","type":"function"},{"doc":"","ref":"Boundary.html#t:call/0","title":"Boundary.call/0","type":"type"},{"doc":"","ref":"Boundary.html#t:dep_error/0","title":"Boundary.dep_error/0","type":"type"},{"doc":"","ref":"Boundary.html#t:error/0","title":"Boundary.error/0","type":"type"},{"doc":"","ref":"Boundary.html#t:mode/0","title":"Boundary.mode/0","type":"type"},{"doc":"","ref":"Boundary.html#t:name/0","title":"Boundary.name/0","type":"type"},{"doc":"","ref":"Boundary.html#t:t/0","title":"Boundary.t/0","type":"type"},{"doc":"","ref":"Boundary.html#t:view/0","title":"Boundary.view/0","type":"type"},{"doc":"Prints information about external dependencies of all application boundaries. Note that :stdlib, :kernel, :elixir, and :boundary will not be included in the output.","ref":"Mix.Tasks.Boundary.FindExternalDeps.html","title":"Mix.Tasks.Boundary.FindExternalDeps","type":"task"},{"doc":"Prints information about all boundaries in the application.","ref":"Mix.Tasks.Boundary.Spec.html","title":"Mix.Tasks.Boundary.Spec","type":"task"},{"doc":"Verifies cross-module function calls according to defined boundaries. This compiler reports all cross-boundary function calls which are not permitted, according to the current definition of boundaries. For details on defining boundaries, see the docs for the Boundary module. Usage Once you have configured the boundaries, you need to include the compiler in mix.exs: defmodule MySystem.MixProject do # ... def project do [ compilers: [:boundary] ++ Mix.compilers(), # ... ] end # ... end When developing a library, it&#39;s advised to use this compiler only in :dev and :test environments: defmodule Boundary.MixProject do # ... def project do [ compilers: extra_compilers(Mix.env()) ++ Mix.compilers(), # ... ] end # ... defp extra_compilers(:prod), do: [] defp extra_compilers(_env), do: [:boundaries] end Warnings Every invalid cross-boundary call is reported as a compiler warning. Consider the following example: defmodule MySystem.User do def auth() do MySystemWeb.Endpoint.url() end end Assuming that calls from MySystem to MySystemWeb are not allowed, you&#39;ll get the following warning: $ mix compile warning: forbidden call to MySystemWeb.Endpoint.url/0 (calls from MySystem to MySystemWeb are not allowed) (call originated from MySystem.User) lib/my_system/user.ex:3 Since the compiler emits warnings, mix compile will still succeed, and you can normally start your system, even if some boundary rules are violated. The compiler doesn&#39;t force you to immediately fix these violations, which is a deliberate decision made to avoid disrupting the development flow. At the same time, it&#39;s worth enforcing boundaries on the CI. This can easily be done by providing the --warnings-as-errors option to mix compile.","ref":"Mix.Tasks.Compile.Boundary.html","title":"Mix.Tasks.Compile.Boundary","type":"task"},{"doc":"","ref":"Mix.Tasks.Compile.Boundary.html#diagnostic/2","title":"Mix.Tasks.Compile.Boundary.diagnostic/2","type":"function"},{"doc":"","ref":"Boundary.Error.html","title":"Boundary.Error","type":"exception"},{"doc":"Boundary Boundary is a library which helps managing and restraining cross-module dependencies in Elixir projects. A few examples of the things you can do with boundary include: Prevent invocations from the context layer to the web layer Prevent invocations from the web layer to internal context modules Prevent usage of Phoenix and Plug in the context layer Limit usage of Ecto in the web layer to only Ecto.Changeset Allow :mix modules to be used only at compile time","ref":"readme.html","title":"Boundary","type":"extras"},{"doc":"This library is experimental, untested, and unstable. Interface might change significantly in the future versions. The code is not well tested or optimized, so you might experience crashes, or significant slowdowns during compilation, especially in larger projects. In addition, be aware that Boundary loads the application and all of its dependencies (recursively) during compilation.","ref":"readme.html#warning","title":"Boundary - Warning","type":"extras"},{"doc":"For a detailed reference see docs for Boundary module and mix compiler.","ref":"readme.html#documentation","title":"Boundary - Documentation","type":"extras"},{"doc":"To use this library, you first need to define the boundaries of your project. A boundary is a named group of one or more modules. Each boundary exports some (but not all!) of its modules, and can depend on other boundaries. During compilation, the boundary compiler will find and report all cross-module function calls which are not permitted according to the boundary configuration. Example Add boundary as a dependency in mix.exs: defmodule MySystem.MixProject do use Mix.Project # ... defp deps do [ {:boundary, &quot;~&gt; 0.4.0&quot;, runtime: false}, # ... ] end # ... end The following code defines boundaries for a typical Phoenix based project generated with mix phx.new. defmodule MySystem do use Boundary, deps: [], exports: [] # ... end defmodule MySystemWeb do use Boundary, deps: [MySystem], exports: [Endpoint] # ... end defmodule MySystem.Application do use Boundary, deps: [MySystem, MySystemWeb] # ... end The configuration above defines three boundaries: MySystem, MySystemWeb, and MySystem.Application. Boundary modules are determined automatically from the boundary name. For example, the MySystem boundary contains the MySystem module, as well as any module whose name starts with MySystem. (e.g. MySystem.User, MySystem.User.Schema, ...). The configuration specifies the following rules: Modules residing in the MySystemWeb boundary are allowed to invoke functions from modules exported by the MySystem boundary. Modules residing in the MySystem.Application namespace are allowed to invoke functions from modules exported by MySystem and MySystemWeb boundaries. All other cross-boundary calls are not permitted. Next, you need to add the mix compiler: defmodule MySystem.MixProject do use Mix.Project def project do [ compilers: [:boundary, :phoenix, :gettext] ++ Mix.compilers(), # ... ] end # ... end Boundary rules are validated during compilation. For example, if we have the following code: defmodule MySystem.User do def auth do MySystemWeb.Endpoint.url() end end The compiler will emit a warning: $ mix compile warning: forbidden call to MySystemWeb.Endpoint.url/0 (calls from MySystem to MySystemWeb are not allowed) (call originated from MySystem.User) lib/my_system/user.ex:3 The complete working example is available here. Because boundary is implemented as a mix compiler, it integrates seamlessly with editors which can work with mix compiler. For example, in VS Code with Elixir LS:","ref":"readme.html#basic-usage","title":"Boundary - Basic usage","type":"extras"},{"doc":"[x] validate calls to external deps (e.g. preventing Ecto usage from MySystemWeb, or Plug usage from MySystem) [x] support compile time vs runtime deps [ ] support nested boundaries (defining internal boundaries within a boundary) [ ] support Erlang modules","ref":"readme.html#roadmap","title":"Boundary - Roadmap","type":"extras"},{"doc":"MIT","ref":"readme.html#license","title":"Boundary - License","type":"extras"}]